//
// This is CS50 Sandbox.
//
// David J. Malan
// malan@harvard.edu
//

// parse command-line arguments
// https://github.com/substack/node-optimist
var argv = require('optimist').usage('Usage: $0 -p port').demand(['p']).argv;

// simplifies asynchronous code
// https://github.com/caolan/async
var async = require('async');

// for spawning childen
var child_process = require('child_process');

// server
var express = require('express');
var app = express();
var server = require('http').createServer(app);
var io = require('socket.io').listen(server);

// for ZIPs
var BufferJoiner = require('bufferjoiner');

// for local files
var fs = require('fs');
var path = require('path');

// for parsing sandbox50.ini
var iniparser = require('iniparser');

// for detecting MIME types
var mime = require('mime');

// jQuery
var $ = require('jquery');

// underscore.js
var _ = require('underscore');

// for UUIDs
var uuid = require('node-uuid');

// for recursive rmdir
var wrench = require('wrench');

// catch any accidentally uncaught exceptions to ensure server's uptime;
// should be replaced with domains once Fedora supports a newer version of node.js,
// per http://nodejs.org/api/process.html#process_event_uncaughtexception
process.on('uncaughtException', function(err) {
    console.log(err);
    console.trace();
});

// load configuration
var config = iniparser.parseSync(path.join(__dirname, 'etc', 'sandbox50.ini'));
for (var key in config) {
    config[key] = $.trim(config[key]);
    if (config[key].length === 0) {
        delete config[key];
    }
}

// decide whether to use S3
var s3cfg = iniparser.parseSync(path.join(__dirname, 'etc', 's3cfg'));
child_process.execFile('s3cmd', ['-c', path.join(__dirname, 'etc', 's3cfg'), 'info', 's3://' + s3cfg.bucket50 + '/'], function(err) {
    if (err === null) {
        console.log('Using S3 for shared storage...');
    }
    else {
        console.log('Not using S3 for shared storage...');
        s3cfg = undefined;
    }
});

// sandboxes, each of the form
// { child: { process: ChildProcess, timeout: object }, homedir: string, timeout: object }
var sandboxes = {};

// set umask for sandboxes, temporaries, uploads
process.umask(0077);

// ensure directory for results exists
wrench.rmdirSyncRecursive(config.RESULTS, true);
wrench.mkdirSyncRecursive(config.RESULTS);
fs.chmodSync(config.RESULTS, 0770);

// ensure directory for temporaries exists
wrench.rmdirSyncRecursive(config.TMP, true);
wrench.mkdirSyncRecursive(config.TMP);
fs.chmodSync(config.TMP, 0700);

// ensure directory for sandboxes exists, but let teardown rmdir
wrench.mkdirSyncRecursive(config.SANDBOXES);
fs.chmodSync(config.SANDBOXES, 0771);

// purge orphaned sandboxes (in case server was restarted before all teardowns fired)
fs.readdir(config.SANDBOXES, function(err, files) {
    if (err === null) {
        async.forEach(files, function(item, callback) {
            if (isId(item)) {
                teardown(item);
            }
        });
    }
});

// purge cached uploads
fs.readdir(config.UPLOADS, function(err, files) {
    if (err === null) {
        async.forEach(files, function(item, callback) {
            if (isId(item)) {
                wrench.rmdirRecursive(path.join(config.UPLOADS, item), function(err) {
                    if (err !== null) {
                        console.log(err.message); // TODO: remove me once sure all's well
                    }
                });
            }
        });
    }
});

// infer server's version from RPM
var VERSION;
child_process.execFile('rpmquery', ['sandbox50'], function(err, stdout, stderr) {
    VERSION = $.trim(stdout);
});

// get server's IP
var ip = (function() {
    try {
        return require('os').networkInterfaces().eth0[0].address;
    }
    catch (e) {
        return undefined;
    }
})();

// configure server
app.configure(function() {

    // parse requests for ZIPs
    // http://steelballsafety.wordpress.com/2012/01/25/express-rawbody/
    app.use(function(req, res, next) {
        if (req.headers['content-type'] === 'application/zip') {

            // store ZIP as a Buffer in .rawBody
            var buffers = new BufferJoiner();
            req.on('data', function(buffer) {
                buffers.add(buffer);
            });
            req.on('end', function() {
                req.rawBody = buffers.join(false);
                next();
            });

        }
        else {
            next();
        }
    });

    // parse requests for other files
    app.use(express.bodyParser({ keepExtensions: true, uploadDir: config.TMP }));

    // limit size of uploads
    app.use(express.limit(config.LIMIT));

    // TODO: delete me once comfortable eliminating SANDBOX50 cookie
    /*
    // inject SANDBOX50 cookie into all responses, in case sandbox50 and a site
    // in which run50 is embedded are both using AWS ELB, lest both sites use
    // default AWSELB cookie, which will break load-balancing for one or the other
    app.use(express.cookieParser());
    app.use(function(req, res, next) {
        if (!isId(req.cookies.SANDBOX50)) {
            res.cookie('SANDBOX50', mkId(), {
             expires: new Date(Date.now() + config.TIMEOUT * 10),
             httpOnly: true
            });
        }
        next();
    });
    */

    // remove X-Powered-By header
    app.use(function(req, res, next) {
        res.removeHeader('x-powered-by');
        next();
    });

    // enable CORS
    app.use(function(req, res, next) {

        // allow clients to send AWSELB cookie in XMLHttpRequests
        res.header('Access-Control-Allow-Credentials', true);

        // whitelist client
        if (!_.isUndefined(req.headers.origin)) {
            res.header('Access-Control-Allow-Origin', req.headers.origin);
        }

        // allow JSON responses to be embedded in DOMs
        res.header('Access-Control-Allow-Methods', 'POST');

        next();
    });

    // route requests
    app.use(app.router);

    // error handling
    // http://expressjs.com/guide.html#error-handling
    app.use(function(err, req, res, next) {
        switch (err.status) {
            case 413:
                return res.json({
                 error: error('E_SIZE')
                }, 400);

            default:
                return res.json({
                 error: error('E_UNKNOWN')
                }, 400);
        }
    });

});

// development settings
app.configure('development', function() {
    app.use(express.errorHandler({ dumpExceptions: true, showStack: true }));
});

// production settings
app.configure('production', function() {
    app.use(express.errorHandler());
});

// disable layouts
app.set('view options', {
 layout: false
});

// disable websocket because it's not supported by Amazon ELB (with stickiness)
io.set('transports', ['jsonp-polling']);

// logging
io.set('log level', 1);

// minify client's JS
io.set('browser client minification', true);

// This fixes issues with terminating the SSL in front of Node and forcing location to think it's wss instead of ws.
// https://github.com/LearnBoost/Socket.IO/wiki/Configuring-Socket.IO
io.set('match origin protocol', true);

// listen for connections on port p
server.listen(argv.p);

// GET /
app.get('/*', function(req, res, next) {

    // get query's keys, if any
    var keys = _.keys(req.query);

    // handle query
    if (keys.length === 1 && _.contains(['check', 'run'], keys[0])) {

        // check
        if (isId(req.query.check)) {

            async.series([

                // check local disk (and S3) for results
                function(callback) {
                    path.exists(path.join(config.RESULTS, req.query.check + '.json'), function(exists) {
                        if (exists === false) {
                            if (!_.isUndefined(s3cfg)) {
                                child_process.execFile('s3cmd', ['-c', path.join(__dirname, 'etc', 's3cfg'), 'get', 's3://' + s3cfg.bucket50 + '/checks/' + req.query.check + '.json', config.RESULTS], function(err, stdout, stderr) {
                                    if (err !== null) {
                                        callback(error('E_USAGE', 'nonexistent results'));
                                    }
                                    else {
                                        callback(null);
                                    }
                                });
                            }
                        }
                        else {
                            callback(null);
                        }
                    });
                }
            ], function(err) {

                if (err !== null) {
                    res.json({
                     error: err
                    }, 400);
                }
                else {
                    fs.readFile(path.join(config.RESULTS, req.query.check + '.json'), function(err, data) {
                        if (err !== null) {
                            res.json({
                             error: err
                            }, 400);
                        }
                        else {
                            res.header('Content-Type', 'text/plain');
                            res.send(data.toString());
                        }
                    });
                }
            });
        }

        // run
        else if (keys[0] === 'run') {
            res.send('coming soon');
        }

    }
    else {

        // map DocumentRoot to ./public
        express.static(__dirname + '/public')(req, res, next);

    }

});

// POST /check
app.post('/check', function(req, res) {
    check(req.body.homedir, req.body.checks, function(err, results) {

        // generate ID for results
        var id = mkId();

        // save results
        async.series([

            // save locally
            function(callback) {
                fs.writeFile(path.join(config.RESULTS, id + '.json'), JSON.stringify(results, null, ' '), callback);
            },

            // save to S3
            function(callback) {
                if (!_.isUndefined(s3cfg)) {
                	child_process.execFile('s3cmd', ['-c', path.join(__dirname, 'etc', 's3cfg'), 'put', '--acl-private', path.join(config.RESULTS, id + '.json'), 's3://' + s3cfg.bucket50 + '/checks/']);
                }
                callback(null);
            }

        ], function(_err) {

            // report results
            if (_err !== null) {
                res.json({
                    error: _err,
                    id: id
                }, 400);
            }
            else if (err !== null) {
                res.json({
                    error: err,
                    id: id
                }, 400);
            }
            else {
                results.id = id;
                res.json(results, 200);
            }

        });

    });
});

// POST /run
app.post('/run', function(req, res) {

    // run cmd in sandbox
    async.waterfall([

        // create sandbox, if not created already
        function(callback) {
            setup(req.body.sandbox, callback);
        },

        // run cmd
        function(sandbox, callback) {
            run(req.body.cmd, sandbox, callback);
        },

        // listen to child
        function(sandbox, child, callback) {

            // get child's start time
            var t0 = (!_.isUndefined(process.hrtime)) ? process.hrtime() : [0,0];

            // buffers
            var script = [], stderr = [], stdout = [];

            // cap lifetime
            var timeout = _setTimeout(function() {
                kill(sandbox, function() {
                    callback(error('E_TIMEOUT'));
                });
            }, config.TIMEOUT);

            // pass stdin, if any, to client
            // TODO: pass on demand via strace hooking, so we can create complete script?
            if (_.isString(req.body.stdin)) {
                child.stdin.write(req.body.stdin);
            }
            child.stdin.end();

            // listen for stderr (from strace), which might contain stdout and/or stderr from sandbox's child
            child.stderr.on('data', function(data) {

                // parse strace
                var syscalls = parse(data.toString());

                // iterate over syscalls
                for (var i = 0; i < syscalls.length; i++) {

                    // handle syscall
                    switch(syscalls[i].fildes) {

                        // buffer stdout
                        case 1:
                            script.push({ type: 'stdout', value: syscalls[i].buf });
                            stdout.push(syscalls[i].buf);
                            break;

                        // buffer stderr
                        case 2:
                            script.push({ type: 'stderr', value: syscalls[i].buf });
                            stderr.push(syscalls[i].buf);
                            break;
                    }
                }

            });

            // exit
            child.on('exit', function(code) {

                // get child's end time
                var t1 = (!_.isUndefined(process.hrtime)) ? process.hrtime() : [0,0];

                // clear timeout
                timeout.clear();

                // check whether child was already SIGKILLed
                if (child.killed === true) {
                    callback(error('E_KILLED'));
                }

                // respond to client
                else {
                    callback(null, {
                     code: code,
                     debug: {
                      ip: ip,
                      version: VERSION
                     },
                     sandbox: sandbox,
                     script: (script.length > 0) ? script : null,
                     stderr: (stderr.length > 0) ? stderr.join('') : null,
                     stdout: (stdout.length > 0) ? stdout.join('') : null,
                     utime: (!_.isUndefined(process.hrtime)) ? t1[0] * 1000000000 + t1[1] - (t0[0] * 1000000000 + t0[1]) : null
                    });
                }
            });
        }
    ],

    // respond to client
    function(err, result) {
        if (err === null) {
            res.json(result, 200);
        }
        else {
            res.json({
             error: err
            }, 400);
        }
    });
});

// POST /save
app.post('/save', function(req, res) {

    // validate filename
    var keys = _.keys(req.body);
    if (keys.length === 0) {
        return res.json({
            error: error('E_USAGE', 'missing filename')
        }, 400);
    }
    else if (keys.length > 1) {
        return res.json({
            error: error('E_USAGE', 'too many filenames')
        }, 400);
    }
    var filename = keys[0];

    // validate content
    var content = req.body[filename];
    if (!_.isString(content) || content.length === 0) {
        return res.json({
            error: error('E_USAGE', 'invalid content')
        }, 400);
    }

    // trigger file download
    res.setHeader('Content-Disposition', 'attachment; filename=' + encodeURIComponent(filename));
    res.setHeader('Content-Type', 'application/octet-stream');
    res.write(content);
    res.end();

});

// GET /status
app.get('/status', function(req, res) {
    res.send(200);
});

// POST /upload
app.post('/upload', function(req, res) {

    // files uploaded
    var files = [];

    // identifier for upload
    var id = mkId();

    // directory for uploaded files
    var dir = path.join(config.UPLOADS, id);

    // series
    async.series([

        // create directory for uploaded files
        function(callback) {
            try {
                fs.mkdir(dir, 0755, callback);
            }
            catch (e) {
                callback(e);
            }
        },

        // save uploaded files
        function(callback) {

            // application/zip
            if (req.headers['content-type'] === 'application/zip') {

                // unzip ZIP
                var zip = path.join(config.TMP, id);
                async.series([

                    function(callback) {
                        fs.writeFile(zip, req.rawBody, callback);
                    },

                    function(callback) {
                        child_process.execFile('unzip', [zip, '-d', dir], callback);
                    },

                    function(callback) {
                        fs.unlink(zip, callback);
                    }

                ], callback);

            }

            // multipart/form-data
            else {
                async.parallel([

                    // save inputs and textareas
                    function(callback) {
                        async.forEach(_.keys(req.body), function(item, callback) {

                            // if string has non-0 size, save it
                            if (req.body[item].length > 0) {
                                fs.writeFile(path.join(dir, item), req.body[item], callback);
                            }
                            else {
                                callback(null);
                            }

                        }, callback);
                    },

                    // save files
                    function(callback) {

                        // if any type=file inputs were POSTed
                        if (_.isObject(req.files)) {

                            // iterate over inputs
                            async.forEach(_.keys(req.files), function(item, callback) {

                                // if file has a non-zero size and non-empty name, save it
                                if (req.files[item].size !== 0 && req.files[item].name !== '') {
                                    fs.rename(req.files[item].path, path.join(dir, item), callback);
                                }
                                else {
                                    callback(null);
                                }

                            }, callback);
                        }
                        else {
                            callback(null);
                        }
                    }

                ], callback);
            }
        },

        // sync to S3
        function(callback) {
            if (!_.isUndefined(s3cfg)) {
                child_process.execFile('s3cmd', ['-c', path.join(__dirname, 'etc', 's3cfg'), 'put', '--acl-private', '-r', dir, 's3://' + s3cfg.bucket50 + '/uploads/'], callback);
            }
            else {
                callback(null);
            }
        },

        // list files at series' end (to avoid race conditions if done asynchronously above as files are written)
        function(callback) {
            var uploads = [];
            wrench.readdirRecursive(dir, function(err, paths) {
                if (err !== null) {
                    callback(err);
                }
                else if (paths !== null) {
                    uploads = uploads.concat(paths);
                }
                else {
                    uploads.sort();
                    callback(null, uploads);
                }
            });
        }
    ],

    // respond to client
    function(err, results) {
        var uploads = results[results.length - 1];
        if (err === null && uploads.length > 0) {
            res.json({
             debug: {
              ip: ip,
              version: VERSION
             },
             id: id,
             uploads: uploads
            });
        }
        else {
            res.json({
             debug: {
              ip: ip,
              version: VERSION
             },
             error: error('E_UNKNOWN', err)
            });
        }
    });
});

// socket.io
io.sockets.on('connection', function(socket) {

    // run
    socket.on('run', function(data) {

        // run cmd in sandbox
        async.waterfall([

            // create sandbox, if not created already
            function(callback) {
                setup(data.sandbox, callback);
            },

            // run cmd
            function(sandbox, callback) {
                run(data.cmd, sandbox, callback);
            },

            // listen to child
            function(sandbox, child, callback) {

                // get child's start time
                var t0 = (!_.isUndefined(process.hrtime)) ? process.hrtime() : [0,0];

                // buffers
                var script = [], stderr = [], stdout = [];

                // TODO: remove this
                // number of writes to stdout and stderr (so that we can SIGKILL potentially infinite loops)
                var writes = 0;

                // queue for emissions
                var emissions = [];

                // timeout for emissions
                var emitter;

                // throttle emissions to no more than once per 500ms
                var emit = function(event, data) {
                    switch (event)
                    {
                        // try to merge back-to-back sterr emissions
                        case 'stderr':

                            if (emissions.length > 0 && emissions[emissions.length - 1].type === 'stderr') {
                                emissions[emissions.length - 1].value += data;
                            }
                            else {
                                emissions.push({ type: 'stderr', value: data });
                            }

                            break;
    
                        // don't merge back-to-back stdin emissions (because client should receive each)
                        case 'stdin':
                            emissions.push({ type: 'stdin', value: undefined });
                            break;

                        // try to merge back-to-back stdout emissions
                        case 'stdout':
                            if (emissions.length > 0 && emissions[emissions.length - 1].type === 'stdout') {
                                emissions[emissions.length - 1].value += data;
                            }
                            else {
                                emissions.push({ type: 'stdout', value: data });
                            }
                            break;

                    }

                    // if we haven't yet defined an emitter
                    if (_.isUndefined(emitter)) {

                        // define emitter
                        emitter = _setTimeout(function() {

                            // emit queue's head
                            var emission = emissions.shift();
                            socket.emit(emission.type, emission.value);

                            // if tail remains, emit again in 500ms
                            if (emissions.length > 0) {
                                emitter.touch();
                            }

                            // else forget timeout
                            else {
                                emitter = undefined;
                            }

                        }, 500);
                    }

                };

                // stderr
                child.stderr.on('data', function(data) {

                    // iterate over syscalls
                    var syscalls = parse(data.toString());
                    for (var i = 0; i < syscalls.length; i++) {

                        // beware excessive output (which overwhelms browsers)
                        if (syscalls[i].fildes == 1 || syscalls[i].fildes == 2) {

                            // TODO: remove this
                            /*
                            if (++writes > 1024) {
                                return kill(sandbox, function(err) {
                                    callback(error('E_EXCESS'));
                                });
                            }
                            */

                        }

                        // handle syscall
                        switch(syscalls[i].fildes) {
                            
                            // inform client of each read on stdin (even though no more than 1 is expected)
                            case 0: 
                                emit('stdin');
                                break;

                            // inform client of stdout
                            case 1: 
                                script.push({ type: 'stdout', value: syscalls[i].buf });
                                stdout.push(syscalls[i].buf);
                                emit('stdout', syscalls[i].buf);
                                break;

                            // inform client of stderr
                            case 2: 
                                script.push({ type: 'stderr', value: syscalls[i].buf });
                                stderr.push(syscalls[i].buf);
                                emit('stderr', syscalls[i].buf);
                                break;
                        }
                    }
                });

                // ctrl-d
                socket.on('EOF', function() {
                    child.stdin.end();
                });

                // kill
                socket.on('SIGINT', function() {
                    kill(sandbox, function(err) {
                        callback(error('E_KILLED'));
                    });
                });

                // stdin
                socket.on('stdin', function(data) {

                    // validate stdin
                    if (!_.isString(data)) {
                        return callback(error('E_UNKNOWN', 'invalid stdin'));
                    }

                    // reset flush counter
                    flushes = 0;

                    // extend lifetime of sandbox's child
                    lifetime.touch();

                    // write stdin to child
                    script.push({ type: 'stdin', value: data });
                    child.stdin.write(data);

                });
 
                // cap lifetime
                var lifetime = _setTimeout(function() {
                    kill(sandbox, function(err) {
                        callback(error('E_TIMEOUT'));
                    });
                }, config.TIMEOUT);

                // exit
                child.on('exit', function(code) {

                    // get child's start time
                    var t1 = (!_.isUndefined(process.hrtime)) ? process.hrtime() : [0,0];

                    // uncap lifetime
                    lifetime.clear();

                    // flush emissions
                    if (!_.isUndefined(emitter)) {
                        emitter.clear();
                    }
                    while (emissions.length > 0) {
                        var emission = emissions.shift();
                        socket.emit(emission.type, emission.value);
                    }

                    // ensure child executed
                    if (code === null) {
                        callback(error('E_UNKNOWN'));
                    }

                    // emit success
                    else {
                        callback(null, {
                         cmd: data.cmd,
                         code: code,
                         debug: {
                          ip: ip,
                          version: VERSION
                         },
                         sandbox: sandbox,
                         script: (script.length > 0) ? script : null,
                         stderr: (stderr.length > 0) ? stderr.join('') : null,
                         stdout: (stdout.length > 0) ? stdout.join('') : null,
                         utime: (!_.isUndefined(process.hrtime)) ? t1[0] * 1000000000 + t1[1] - (t0[0] * 1000000000 + t0[1]) : null
                        });
                    }

                });

           }],

            // handle error
            function(err, result) {
                if (err === null) {
                    socket.emit('exit', result, 200);
                }
                else {
                    socket.emit('error', {
                        error: err
                    }, 400);
                }
            }
        );
    });
});

/**
 * Wraps setTimeout so that caller can extend timeout without having to remember handler.
 *
 * @param Function func
 * @param number delay
 */
function _setTimeout(func, delay) {
    return {
     clear: function() { 
        if (!_.isUndefined(this.timeout)) {
            clearTimeout(this.timeout);
            this.timeout = undefined;
        }
     },
     func: func,
     timeout: setTimeout(func, delay),
     touch: function() { 
        if (!_.isUndefined(this.timeout)) {
            clearTimeout(this.timeout);
            this.timeout = setTimeout(func, delay);
        }
     }
    };
}

/**
 * Checks homedir.
 *
 * Inspired by Caolan McMahon's async.auto (https://github.com/caolan/async/#auto).
 *
 * @param string homedir
 * @param string checks
 * @param function callback
 */
function check(homedir, checks, callback) {

    // validate homedir
    if (_.isUndefined(homedir)) {
        return callback(error('E_USAGE', 'missing homedir'));
    }
    else if (!isId(homedir)) {
        return callback(error('E_USAGE', 'invalid homedir'));
    }

    // validate checks
    if (_.isUndefined(checks)) {
        return callback(error('E_USAGE', 'missing checks'));
    }
    else if (!_.isString(checks)) {
        return callback(error('E_USAGE', 'invalid checks'));
    }

    // TODO: change foo/bar/baz format to foo.bar.baz or similar?
    // normalize path to ensure checks are confined to expected directory
    var module = path.join(config.CHECKS, checks);
    if (!module.match(new RegExp('^' + config.CHECKS + '/'))) {
        return callback(error('E_USAGE', 'illegal checks'));
    }

    // purge module from cache in development mode, lest checks have changed
    if (app.settings.env === 'development') {
        delete require.cache[require.resolve(module)];
    }

    // import checks
    try {
        checks = require(module);
    }
    catch (e) {
        return callback(error('E_USAGE', 'non-existent checks'));
    }

    // get check's keys
    var keys = _.keys(checks);

    // if no keys, no checks!
    if (keys.length === 0) {
        return callback(null, { results: null });
    }

    // initialize results' values to null so that keys' order is maintained
    // (unless numeric, per http://stackoverflow.com/a/3186907)
    var results = {};
    for (key in checks) {
        results[key] = null;
    }

    // listeners for checking dependencies' satisfaction
    var listeners = [];
    var addListener = function(fn) {
        listeners.unshift(fn);
    };
    var removeListener = function(fn) {
        for (var i = 0; i < listeners.length; i++) {
            if (listeners[i] === fn) {
                listeners.splice(i, 1);
                return;
            }
        }
    };

    // respond to caller once all checks have results
    addListener(function () {
        if (!_.any(_.values(results), _.isNull)) {

            // prevent multiple responses
            callback(null, { results: results });
            callback = function () {};

            // TODO: decide if necessary
            // remove listeners
            //listeners = [];
        }
    });

    // iterate over check's keys
    _.each(keys, function(key) {

        // get check's value, which is either a function or an array of the
        // form [d_1, d_2, d_n, f], where f is a function and each d_i is a
        // dependency (i.e., another check that must be executed first)
        var value = (_.isFunction(checks[key])) ? [checks[key]]: checks[key];

        // determine check's requirements
        var dependencies = value.slice(0, Math.abs(value.length - 1)) || [];

        // TODO: check for circular dependencies (whereby foo is a dependency of foo or
        // whereby foo is a dep of bar, and bar is a dep of foo)

        // determines whether check's dependencies are satisified
        var satisfied = function() {
            return (_.reduce(dependencies, function(a, x) {
                return (a && !_.isNull(results[x]));
            }, true) && _.isNull(results[key]));
        };

        // function that will be passed to checks with which they can start a chain
        var check = function(description) {

            // each check is allowed one child at a time in a sandbox
            var child, sandbox;

            // timeout for any .run links
            var timeout;

            // callback for when all of a chain's links have been processed
            var callback = function(err, result) {

                // clear any timeout
                if (_.isObject(timeout)) {
                    timeout.clear();
                }

                // tear down sandbox
                teardown(sandbox);

                // record error
                if (err !== null) {
                    results[key] = {
                     dependencies: dependencies,
                     description: description,
                     error: err.message
                    };
                }

                // or result
                else {
                    results[key] = {
                     dependencies: dependencies,
                     description: description,
                     result: result,
                     script: script
                    };
                }

                // lest another call to chain() be scheduled for the next tick,
                // be sure we can't call this callback again for this key
                callback = function() {};

                // let other checks know that this check is complete (in case it's a dependency of theirs)
                _.each(listeners.slice(0), function(fn) {
                    fn();
                });
            };

            // buffers for any children's stderr and stdout (so that regexes match greedily)
            var stderr = [], stdout = [];

            // queues
            var events = [], links = [];

            // lock
            var locked = false;

            // script
            var script = [];

            // chain
            // inspired by https://github.com/caolan/async#auto
            var chain = function(s) {

                // configure chain for sandbox
                if (arguments.length == 1) {
                
                    // ensure sandbox is valid
                    if (!isId(s)) {
                        return callback(error('E_UNKNOWN', 'invalid sandbox'));
                    }

                    // ensure chain wasn't already configured
                    if (!_.isUndefined(sandbox)) {
                        return callback(error('E_UNKNOWN', 'chain already configured'));
                    }

                    // check whether dependencies passed, failed, or erred before processing chain
                    for (var i = 0; i < dependencies.length; i++) {
                        if (!_.isUndefined(results[dependencies[i]].error)) {
                            return callback(error('E_CONFIG', 'error in dependency'));
                        }
                        else if (results[dependencies[i]].result !== true) {
                            return callback(null, null);
                        }
                    }

                    // validate sandbox
                    if (!isId(s)) {
                        return callback(error('E_UNKNOWN', 'invalid sandbox'));
                    }

                    // remember sandbox
                    sandbox = s;
                }
                else if (arguments.length !== 0) {
                    return callback(error('E_CONFIG', 'invalid configuration of chain'));
                }

                // check whether chain's already been processed,
                // in case some event queued this otherwise extra call via setTimeout
                if (!_.isNull(results[key])) {
                    return;
                }

                // check lock
                if (locked === true) {
                    return;
                }

                // if out of links, check passed!
                if (links.length === 0) {
                    return callback(null, true);
                }

                // process link in chain
                var link = links.shift();

                // process event, if any
                var event = (events.length > 0) ? events.shift() : undefined;

                // check for SIGKILL
                if (!_.isUndefined(event) && event.type === 'signal' && event.value === 'SIGKILL') {

                    // script this
                    script.push({
                     actual: event
                    });

                    return callback(null, false);
                }

                // .cp
                if (link.fn === 'cp') {

                    // postpone handling any event
                    if (!_.isUndefined(event)) {
                        events.unshift(event);
                    }

                    // normalize src to ensure it's confined to expected directory
                    var src = path.join(module, link.args[0]);
                    if (!src.match(new RegExp('^' + module + '/'))) {
                        return callback(error('E_USAGE', 'illegal file in check'));
                    }

                    // TODO: make homedir and tmpdir properties of a Sandbox object?
                    var homedir = path.join(config.SANDBOXES, sandbox, 'homedir');

                    // normalize dst to ensure it's confined to expected directory
                    var dst = path.join(homedir, link.args[1]);
                    if (!dst.match(new RegExp('^' + homedir + '/'))) {
                        return callback(error('E_USAGE', 'illegal path in check'));
                    }

                    // script this
                    script.push({
                     expected: { type: link.fn, value: link.args[0] }
                    });

                    // lock chain
                    locked = true;

                    // cp -rf src dst 
                    async.series([
                        function(callback) {
                            child_process.execFile('sudo', ['cp', '-rf', src, dst], callback);
                        },
                        function(callback) {
                            child_process.execFile('sudo', ['chown', '-R', [sandbox, ':', 'sandboxes'].join(''), dst], callback);
                        }
                    ], function(err) {

                        // unlock chain
                        locked = false;

                        // check for error
                        if (err !== null) {
                            return callback(err);
                        }

                        // try to process another link on next tick
                        setTimeout(function() { chain(); }, 0);

                    });

                }

                // .diff
                else if (link.fn === 'diff') {

                    // postpone handling any event
                    if (!_.isUndefined(event)) {
                        events.unshift(event);
                    }

                    // normalize expected to ensure it's confined to expected directory
                    var expected = path.join(module, link.args[0]);
                    if (!expected.match(new RegExp('^' + module + '/'))) {
                        return callback(error('E_USAGE', 'illegal file in check'));
                    }

                    // TODO: make homedir and tmpdir properties of a Sandbox object?
                    var homedir = path.join(config.SANDBOXES, sandbox, 'homedir');

                    // normalize actual to ensure it's confined to expected directory
                    var actual = path.join(homedir, link.args[1]);
                    if (!actual.match(new RegExp('^' + homedir + '/'))) {
                        return callback(error('E_USAGE', 'illegal path in check'));
                    }

                    // lock chain
                    locked = true;

                    // parallelize for efficiency
                    async.parallel([

                        // sudo diff -q expected actual
                        function(callback) {
                            child_process.execFile('sudo', ['diff', '-q', expected, actual], function(err, stdout, stderr) {
                                callback(null, (_.isNull(err)) ? 0 : err.code);
                            });
                        },

                        // fs.readFile expected
                        function(callback) {
                            fs.readFile(expected, callback);
                        },

                        // sudo cat actual
                        function(callback) {
                            child_process.execFile('sudo', ['cat', actual], function(err, stdout, stderr) {
                                callback(err, stdout);
                            });
                        }

                    ], function(err, results) {

                        // unlock chain
                        locked = false;

                        // check for error
                        if (err !== null) {
                            return callback(err);
                        }

                        // script this, for text files
                        if (mime.lookup(expected).match(/text/) && mime.lookup(actual).match(/text/)) {
                            script.push({
                             actual: { type: 'diff', value: results[2] },
                             expected: { type: link.fn, value: results[1].toString() }
                            });
                        }

                        // script this, for binary files
                        else {

                            // convert Buffers to Arrays of Numbers
                            script.push({
                             actual: { type: 'diff', value: _.map(new Buffer(results[2], 'binary'), function(d) { return d; }) },
                             expected: { type: link.fn, value: _.map(results[1], function(d) { return d; }) }
                            });

                        }

                        // check failed if files differ
                        if (results[0] !== 0) {
                            return callback(null, false);
                        }

                        // try to process another link on next tick
                        setTimeout(function() { chain(); }, 0);

                    });

                }

                // .exists link
                else if (link.fn === 'exists') {

                    // postpone handling any event
                    if (!_.isUndefined(event)) {
                        events.unshift(event);
                    }

                    // TODO: make homedir and tmpdir properties of a Sandbox object?
                    var homedir = path.join(config.SANDBOXES, sandbox, 'homedir');

                    // TODO: support an array of filenames (e.g., ['resize.c', 'bmp.h'])

                    // normalize filename to ensure it's confined to expected directory
                    var filename = path.join(homedir, link.args[0]);
                    if (!filename.match(new RegExp('^' + homedir + '/'))) {
                        return _callback(error('E_USAGE', 'illegal path in check'));
                    }

                    // script this
                    script.push({
                     expected: { type: link.fn, value: link.args[0] }
                    });

                    // lock chain
                    locked = true;

                    // check whether file exists
                    path.exists(filename, function(exists) {

                        // unlock chain
                        locked = false;

                        // file should exist
                        if (exists !== true) {
                            return callback(null, false);
                        }

                        // try to process another link on next tick
                        setTimeout(function() { chain(); }, 0);

                    });
                }

                // .exit
                else if (link.fn === 'exit') {

                    // if an event hasn't yet been emitted, postpone handling this link
                    if (_.isUndefined(event)) {
                        links.unshift(link);   
                    }

                    else {

                        // script this
                        script.push({
                         actual: event,
                         expected: { type: link.fn, value: link.args[0] }
                        });

                        // compare expected and actual exit code if we're reacting to an exit event
                        if (event.type !== 'exit' || link.args[0] !== event.value) {
                            return callback(null, false);
                        }

                        // TODO: catch error whereby there's two .exit links in a row

                        // try to process another link on next tick
                        setTimeout(function() { chain(); }, 0);
                    }
                }

                // .patch
                else if (link.fn === 'patch') {

                    // postpone handling any event
                    if (!_.isUndefined(event)) {
                        events.unshift(event);
                    }

                    // TODO: make homedir and tmpdir properties of a Sandbox object?
                    var homedir = path.join(config.SANDBOXES, sandbox, 'homedir');

                    // normalize originalfile to ensure it's confined to expected directory
                    var originalfile = path.join(homedir, link.args[0]);
                    if (!originalfile.match(new RegExp('^' + homedir + '/'))) {
                        return callback(error('E_USAGE', 'illegal path in check'));
                    }

                    // normalize patchfile to ensure it's confined to expected directory
                    var patchfile = path.join(module, link.args[1]);
                    if (!patchfile.match(new RegExp('^' + module + '/'))) {
                        return callback(error('E_USAGE', 'illegal file in check'));
                    }

                    // script this
                    script.push({
                     expected: { type: link.fn, value: link.args[0] }
                    });

                    // lock chain
                    locked = true;

                    // patch -l -u originalfile patchfile
                    async.series([
                        function(callback) {
                            child_process.execFile('sudo', ['patch', '-l', '-u', originalfile, patchfile], callback);
                        },
                        function(callback) {
                            child_process.execFile('sudo', ['chown', '-R', [sandbox, ':', 'sandboxes'].join(''), originalfile], callback);
                        }
                    ], function(err) {

                        // unlock chain
                        locked = false;

                        // check for error
                        if (err !== null) {
                            return callback(new Error('could not patch ' + link.args[0]));
                        }

                        // try to process another link on next tick
                        setTimeout(function() { chain(); }, 0);

                    });

                }


                // .run link
                else if (link.fn === 'run') {

                    // script this
                    script.push({
                     actual: event,
                     expected: { type: link.fn, value: link.args[0] }
                    });

                    // err if one or more events are still outstanding
                    if (!_.isUndefined(event)) {
                        return callback(null, false);
                    }

                    // lock chain
                    locked = true;

                    // run cmd
                    run(link.args[0], sandbox, function(err, _sandbox, _child) {

                        // unlock chain
                        locked = false;

                        // check for error
                        if (err !== null) {
                            return callback(err);
                        }

                        // cap lifetime
                        timeout = _setTimeout(function() {

                            // queue event
                            events.push({ type: 'signal', value: 'SIGKILL' })

                            // forget child
                            child = undefined;

                            // kill sandbox
                            kill(sandbox);

                            // try to process another link on next tick (if only to handle this event)
                            setTimeout(function() { chain(); }, 0);

                        }, config.TIMEOUT);

                        // remember child
                        child = _child;

                        // listen for exit
                        child.on('exit', function(code) {

                            // clear timeout
                            timeout.clear();

                            // queue event
                            events.push({ type: 'exit', value: code });

                            // forget child
                            child = undefined;

                            // try to process another link on next tick (if only to handle this event)
                            setTimeout(function() { chain(); }, 0);

                        });

                        // listen for stderr from strace
                        child.stderr.on('data', function(data) {

                            // parse syscalls
                            var syscalls = parse(data.toString());

                            // number of events queued before these syscalls
                            var length = events.length;

                            // iterate over syscalls, consolidating like events to maximize values' lengths
                            // (so that caller sees as much as possible per event)
                            for (var i = 0; i < syscalls.length; i++) {
                                switch (syscalls[i].fildes) {

                                    // stdin
                                    case 0:
                                        events.push({ type: 'stdin' });
                                        break;

                                    // stdout
                                    case 1:
                                        if (events.length > 0 && events[events.length - 1].type === 'stdout') {
                                            events[events.length - 1].value += syscalls[i].buf;
                                        }
                                        else {
                                            events.push({ type: 'stdout', value: syscalls[i].buf });
                                        }
                                        break;

                                    // stderr
                                    case 2:
                                        if (events.length > 0 && events[events.length - 1].type === 'stderr') {
                                            events[events.length - 1].value += syscalls[i].buf;
                                        }
                                        else {
                                            events.push({ type: 'stderr', value: syscalls[i].buf });
                                        }
                                        break;
                                }
                            }

                            // TODO: decide whether comparing length before and after is sufficient
                            // try to process another link on next tick
                            if (events.length > length) {
                                setTimeout(function() { chain(); }, 200); // wait a bit for more stderr or stdout
                            }

                        });

                        // try to process another link on next tick
                        setTimeout(function() { chain(); }, 0);

                    });
                }

                // .signal link + signal event
                else if (link.fn === 'signal') {

                    // TODO: add support for signals
                    return callback(error('E_USAGE', '.signal is not yet supported'));

                }

                // .stderr
                else if (link.fn === 'stderr') {
                
                    // if an event hasn't yet been emitted, postpone handling this link
                    if (_.isUndefined(event)) {
                        links.unshift(link);
                    }

                    // if a stderr event was emitted
                    else if (event.type === 'stderr') {

                        // buffer stderr in case there's more to come
                        // so that this link's regex can match greedily
                        stderr.push(event.value);

                        // and postpone processing this link
                        links.unshift(link);

                        // try to process another link on next tick
                        setTimeout(function() { chain(); }, 0);

                    }

                    // if an exit or stdin event was emitted,
                    // it's time to compare expected and actual stderr
                    else if (event.type === 'exit' || event.type === 'stdin') {

                        // lock chain
                        locked = true;

                        // iterate over link's expected stderr
                        async.map(link.args[0], function(stderr, _callback) {

                            // if stderr is a regex
                            if (stderr instanceof RegExp) {

                                // replace leading / with ( and trailing / with ) so that we can 
                                // eventually add support for case-insensitivity using capturing parentheses
                                return _callback(null, stderr.toString().replace(/^\//, '(').replace(/\/$/, ')'));

                            }

                            // else if stderr is a string, assume it's a filename
                            else if (_.isString(stderr)) {

                                // normalize filename to ensure it's confined to expected directory
                                var filename = path.join(module, stdout);
                                if (!filename.match(new RegExp('^' + module + '/'))) {
                                    return _callback(error('E_USAGE', 'illegal file in check'));
                                }

                                // read file into a string
                                fs.readFile(filename, undefined, function(err, data) {

                                    // surround string with capturing parentheses (for future case sensitivity)
                                    if (err === null) {
                                        _callback(null, '(' + data.toString() + ')');
                                    }
                                    else {
                                        _callback(err);
                                    }
                                });

                            }

                        }, function(err, results) {

                            // unlock chain
                            locked = false;

                            // check for error
                            if (err !== null) {
                                return callback(err);
                            }

                            // construct regex of the form ^(...)(...)...$,
                            // where each (...) is from a .stderr link
                            var expected = new RegExp('^' + results.join('') + '$');

                            // concatenate buffer of actual stderr
                            var actual = stderr.join('');

                            // if we haven't buffered any stderr, can't compare with expectations
                            if (stderr.length === 0) {

                                // script this
                                script.push({
                                 actual: event, 
                                 expected: { type: link.fn, value: expected.toString() }
                                });

                                // we expected stderr
                                return callback(null, false);

                            }

                            // postpone handling this event
                            events.unshift(event);

                            // script this
                            script.push({
                             actual: { type: link.fn, value: actual },
                             expected: { type: link.fn, value: expected.toString() }
                            });

                            // compare expected regex against actual (concatenated) buffer
                            if (expected.test(actual) !== true) {
                                return callback(null, false);
                            }

                            // flush buffer
                            stderr.length = 0;

                            // try to process another link on next tick
                            setTimeout(function() { chain(); }, 0);

                        });
                    }

                    // else unexpected event
                    else {
                        script.push({
                         actual: event,
                         expected: { type: link.fn, value: link.args[0] }
                        });
                        return callback(null, false);
                    }
                }

                // .stdin link
                else if (link.fn === 'stdin') {

                    // TODO: doesn't seem to handle two .stdin events in a row properly yet?

                    // if an event hasn't yet been emitted, postpone handling this link
                    if (_.isUndefined(event)) {
                        links.unshift(link);
                    }

                    // if a stdin event was emitted
                    else if (event.type === 'stdin') {

                        // script this
                        script.push({
                         actual: { type: event.type },
                         expected: { type: link.fn, value: link.args[0] }
                        });

                        // treat null as EOF
                        if (_.isNull(link.args[0])) {
                            child.stdin.end();
                        }

                        // write stdin, if provided, into child
                        else if (!_.isUndefined(link.args[0])) {
                            child.stdin.write(link.args[0]);
                        }

                        // try to process another link on next tick
                        setTimeout(function() { chain(); }, 0);
                    }

                    // else unexpected event
                    else {

                        script.push({
                         actual: event,
                         expected: { type: link.fn, value: link.args[0] }
                        });
                        return callback(null, false);
                    }

                }

                // .stdout
                else if (link.fn === 'stdout') {
                
                    // if an event hasn't yet been emitted, postpone handling this link
                    if (_.isUndefined(event)) {
                        links.unshift(link);
                    }

                    // if a stdout event was emitted
                    else if (event.type === 'stdout') {

                        // buffer stdout in case there's more to come
                        // so that this link's regex can match greedily
                        stdout.push(event.value);

                        // and postpone processing this link
                        links.unshift(link);

                        // try to process another link on next tick
                        setTimeout(function() { chain(); }, 0);

                    }

                    // if an exit or stdin event was emitted,
                    // it's time to compare expected and actual stdout
                    else if (event.type === 'exit' || event.type === 'stdin') {

                        // lock chain
                        locked = true;

                        // iterate over link's expected stdout
                        async.map(link.args[0], function(stdout, _callback) {

                            // if stdout is a regex
                            if (stdout instanceof RegExp) {

                                // replace leading / with ( and trailing / with ) so that we can 
                                // eventually add support for case-insensitivity using capturing parentheses
                                return _callback(null, stdout.toString().replace(/^\//, '(').replace(/\/$/, ')'));

                            }

                            // else if stdout is a string, assume it's a filename
                            else if (_.isString(stdout)) {

                                // normalize filename to ensure it's confined to expected directory
                                var filename = path.join(module, stdout);
                                if (!filename.match(new RegExp('^' + module + '/'))) {
                                    return _callback(error('E_USAGE', 'illegal file in check'));
                                }

                                // read file into a string
                                fs.readFile(filename, undefined, function(err, data) {

                                    // surround string with capturing parentheses (for future case sensitivity)
                                    if (err === null) {

                                        // escape file's contents so that they're not treated as a regex
                                        // http://stackoverflow.com/a/3561711
                                        _callback(null, '(' + data.toString().replace(/[-\/\\^$*+?.()|[\]{}]/g, '\\$&') + ')');
                                    }
                                    else {
                                        _callback(error('E_USAGE', 'invalid check'));
                                    }

                                });

                            }

                        }, function(err, results) {

                            // unlock chain
                            locked = false;

                            // check for error
                            if (err !== null) {
                                return callback(err);
                            }

                            // construct regex of the form ^(...)(...)...$,
                            // where each (...) is from a .stdout link
                            var expected = new RegExp('^' + results.join('') + '$');

                            // concatenate buffer of actual stdout
                            var actual = stdout.join('');

                            // if we haven't buffered any stdout, can't compare with expectations
                            if (stdout.length === 0) {

                                // script this
                                script.push({
                                 actual: event,
                                 expected: { type: link.fn, value: expected.toString() }
                                });

                                // we expected stdout
                                return callback(null, false);

                            }

                            // postpone handling this event
                            events.unshift(event);

                            // script this
                            script.push({
                             actual: { type: link.fn, value: actual },
                             expected: { type: link.fn, value: expected.toString() }
                            });

                            // compare expected regex against actual (concatenated) buffer
                            if (expected.test(actual) !== true) {
                                return callback(null, false);
                            }

                            // flush buffer
                            stdout.length = 0;

                            // try to process another link on next tick
                            setTimeout(function() { chain(); }, 0);

                        });
                    }

                    // else unexpected event
                    else {
                        script.push({
                         actual: event,
                         expected: { type: link.fn, value: link.args[0] }
                        });
                        return callback(null, false);
                    }
 
                }

                // else unexpected link
                else {
                    return callback(error('E_CONFIG', 'illegal link in chain'));
                }

            };

            // returns true iff chain contains a .run that isn't (yet) accompanied by a .exit
            var running = function() {

                // indices of prior links in chain
                var indices = _.range(0, links.length).reverse();

                // find index of most recent .run, if any
                var run = _.find(indices, function(i) {
                    return links[i].fn === 'run';
                });

                // find index of most recent .exit, if any
                var exit = _.find(indices, function(i) {
                    return links[i].fn === 'exit';
                });

                // confirm that chain's tail is preceded by a .run (and not another .exit)
                if (_.isUndefined(run) || (!_.isUndefined(exit) && exit > run)) {
                    return false;
                }
                else {
                    return true;
                }
            };
    
            // .cp
            chain.cp = function(src, dst) {

                // validate argument count
                if (arguments.length !== 2) {
                    callback(error('E_CONFIG', 'invalid number of arguments'));
                }

                // validate src 
                if (!_.isString(src)) {
                    callback(error('E_CONFIG', 'invalid filename'));
                }

                // validate dst
                if (!_.isString(dst)) {
                    callback(error('E_CONFIG', 'invalid filename'));
                }

                // add link to chain
                links.push({
                 args: [src, dst],
                 fn: 'cp'
                });

                // return chain
                return this;
            };

            // .diff
            chain.diff = function(expected, actual) {

                // validate argument count
                if (arguments.length !== 2) {
                    callback(error('E_CONFIG', 'invalid number of arguments'));
                }

                // validate expected
                if (!_.isString(expected)) {
                    callback(error('E_CONFIG', 'invalid filename'));
                }

                // validate actual
                if (!_.isString(actual)) {
                    callback(error('E_CONFIG', 'invalid filename'));
                }

                // add link to chain
                links.push({
                 args: [expected, actual],
                 fn: 'diff'
                });

                // return chain
                return this;
            };

            // .exists
            chain.exists = function(expected) {

                // validate argument count
                if (arguments.length !== 1) {
                    callback(error('E_CONFIG', 'invalid number of arguments'));
                }

                // validate expected
                if (!_.isString(expected)) {
                    callback(error('E_CONFIG', 'invalid filename'));
                }

                // add link to chain
                links.push({
                 args: [expected],
                 fn: 'exists'
                });

                // return chain
                return this;
            };

            // .exit
            chain.exit = function(code) {

                // validate argument count
                if (arguments.length > 1) {
                    callback(error('E_CONFIG', 'invalid number of arguments'));
                }

                // validate code
                if (arguments.length !== 0) {
                    if (!_.isNumber(code) || code % 1 !== 0) { // ensure code is an int
                        callback(error('E_CONFIG', 'invalid code'));
                    }
                }

                // validate placement of .exit in chain
                if (!running()) {
                    callback(error('E_CONFIG', 'missing .run before .exit'));
                }

                // add link to chain
                links.push({
                 args: (arguments.length === 1) ? [code] : [],
                 fn: 'exit'
                });

                // return chain
                return this;
            };

            // .patch
            chain.patch = function(originalfile, patchfile) {

                // validate argument count
                if (arguments.length !== 2) {
                    callback(error('E_CONFIG', 'invalid number of arguments'));
                }

                // validate originalfile
                if (!_.isString(originalfile)) {
                    callback(error('E_CONFIG', 'invalid filename'));
                }

                // validate patchfile
                if (!_.isString(patchfile)) {
                    callback(error('E_CONFIG', 'invalid filename'));
                }

                // add link to chain
                links.push({
                 args: [originalfile, patchfile],
                 fn: 'patch'
                });

                // return chain
                return this;
            };

            // .run
            chain.run = function(cmd) {

                // validate argument count
                if (arguments.length !== 1) {
                    callback(error('E_CONFIG', 'invalid number of arguments'));
                }

                // validate cmd
                if (!_.isString(cmd)) { 
                    callback(error('E_CONFIG', 'invalid cmd'));
                }

                // validate placement of .run in chain
                if (running()) {
                    callback(error('E_CONFIG', 'missing .exit before .run'));
                }

                // add link to chain
                links.push({
                 args: [cmd],
                 fn: 'run'
                });

                // return chain
                return this;
            };

            // .signal
            chain.signal = function(signal) {

                // validate argument count
                if (arguments.length !== 1) {
                    callback(error('E_CONFIG', 'invalid number of arguments'));
                }

                // validate signal
                if (!_.isNumber(signal) || signal % 1 !== 0) { // ensure signal is an int
                    callback(error('E_CONFIG', 'invalid signal'));
                }

                // validate placement of .signal in chain
                if (!running()) {
                    callback(error('E_CONFIG', 'missing .run before .signal'));
                }

                // add link to chain
                links.push({
                 args: [signal],
                 fn: 'signal'
                });

                // return chain
                return this;
            };

            // .stderr
            chain.stderr = function(stderr) {

                // validate argument count
                if (arguments.length > 1) {
                    callback(error('E_CONFIG', 'invalid number of arguments'));
                }

                // .stderr() shall be shorthand for .stderr(/[\s\S]*/)
                if (_.isUndefined(stderr)) {
                    stderr = new RegExp('[\\s\\S]*');
                }

                // validate stderr
                if (!_.isString(stderr) && !_.isRegExp(stderr)) {
                    callback(error('E_CONFIG', 'invalid stderr'));
                }

                // validate placement of .signal in chain
                if (!running()) {
                    callback(error('E_CONFIG', 'missing .run before .stderr'));
                }

                // if previous link was also stderr, consolidate regexes
                if (links.length > 0 && links[links.length - 1].fn === 'stderr') {
                    links[links.length - 1].args[0].push(stderr);
                }
                else {

                    // else append new link
                    links.push({
                     args: [[stderr]],
                     fn: 'stderr'
                    });
                }

                // return chain
                return this;
            };

            // .stdin
            chain.stdin = function(stdin) {

                // validate argument count
                if (arguments.length > 1) {
                    return callback(error('E_CONFIG', 'invalid number of arguments'));
                }

                // validate stdin
                if (!_.isUndefined(stdin) && !_.isNull(stdin) && !_.isString(stdin)) {
                    return callback(error('E_CONFIG', 'invalid stdin'));
                }

                // validate placement of .stdin in chain
                if (!running()) {
                    callback(error('E_CONFIG', 'missing .run before .stdin'));
                }

                // add link to chain
                links.push({
                 args: [stdin],
                 fn: 'stdin'
                });

                // return chain
                return this;
            };

            // .stdout
            chain.stdout = function(stdout) {

                // validate argument count
                if (arguments.length > 1) {
                    callback(error('E_CONFIG', 'invalid number of arguments'));
                }

                // .stdout() shall be shorthand for .stdout(/[\s\S]*/)
                if (_.isUndefined(stdout)) {
                    stdout = new RegExp('[\\s\\S]*');
                }

                // validate stdout
                if (!_.isString(stdout) && !_.isRegExp(stdout)) {
                    callback(error('E_CONFIG', 'invalid stdout'));
                }

                // validate placement of .stdout in chain
                if (!running()) {
                    callback(error('E_CONFIG', 'missing .run before .stdout'));
                }

                // if previous link was also stdout, consolidate regexes
                if (links.length > 0 && links[links.length - 1].fn === 'stdout') {
                    links[links.length - 1].args[0].push(stdout);
                }

                // else append new link
                else {
                    links.push({
                     args: [[stdout]],
                     fn: 'stdout'
                    });
                }

                // return chain
                return this;
            };

            // return (initial link in) chain
            return chain;
        };

        // set up sandbox for chain
        // TODO: don't set up sandbox until dependencies are satisified
        setup({ homedir: homedir }, function(err, sandbox) {

            // check for erro
            if (err !== null) {
                return callback(err);
            }

            // if check's dependencies are satisfied, process chain
            if (satisfied()) {
                var chain = value[value.length - 1](check);
                chain(sandbox);
            }

            // else listen for dependencies' satisfication, then process chain
            else {
                var listener = function() {
                    if (satisfied()) {

                        // no need to listen anymore
                        removeListener(listener);

                        // process chain
                        var chain = value[value.length - 1](check);
                        chain(sandbox);

                    }
                };
                addListener(listener);
            }
        });
    });
}

/**
 * Returns an error object with a code and messagd.
 *
 * @param string code
 * @param string message
 *
 * @return object
 */
function error(code, message) {
    return { code: code, message: message };
}

/**
 * Returns true iff s is an ID.
 *
 * @param string s
 *
 * @return bool
 */
function isId(s) {
    return _.isString(s) && s.match(/^[a-f0-9]{8}[a-f0-9]{4}4[a-f0-9]{3}[89ab][a-f0-9]{3}[a-f0-9]{12}$/);
}

/**
 * Kills a sandbox's process(es).  Calls callback when done.
 *
 * @param string sandbox
 * @param function callback
 */
function kill(sandbox, callback) {

    // validate sandbox
    if (!_.isString(sandbox) || !_.isObject(sandboxes[sandbox])) {
        if (_.isFunction(callback)) {
            callback(error('E_USAGE', 'invalid sandbox'));
        }
        return;
    }

    // kill sandbox's process(es)
    async.series([

        // killall -STOP -u sandbox
        function(callback) {
            child_process.execFile('sudo', ['killall', '-STOP', '-u', sandbox], function(err, stdout, stderr) {
                callback(null); // don't worry if no processes were stopped (in which case killall's exit code would be non-0)
            });
        },

        // killall -KILL -u sandbox
        function(callback) {
            child_process.execFile('sudo', ['killall', '-KILL', '-u', sandbox], function(err, stdout, stderr) {
                callback(null); // don't worry if no processes were killed (in which case killall's exit code would be non-0)
            });
        },

        // kill -s KILL sandboxes[sandbox].child
        function(callback) {
            if (_.isObject(sandboxes[sandbox].child) && sandboxes[sandbox].child.killed === false && _.isNumber(sandboxes[sandbox].child.pid)) {

                sandboxes[sandbox].child.removeAllListeners(); // so that .exit handler doesn't get called
                child_process.execFile('sudo', ['kill', '-s', 'KILL', sandboxes[sandbox].child.pid], function(err, stdout, stderr) {
                    delete sandboxes[sandbox].child;
                    callback(null); // don't worry if no process was killed (in which case kill's exit code would be non-0)
                });

            }
            else {
                delete sandboxes[sandbox].child;
                callback(null);
            }
        }

    ], function(err) {
        if (_.isFunction(callback)) {
            callback(err);
        }
    });
}

/**
 * Returns a unique identifier (which happens to be a version 4 UUID with lowercase letters and without hyphens).
 *
 * @see http://en.wikipedia.org/wiki/Universally_unique_identifier#Version_4_.28random.29
 *
 * @return string
 */
function mkId() {
    return uuid.v4().toString().replace(/-/g, '').toLowerCase();
}

/**
 * Parses strace's output, returning an array of the form
 * [ { fildes: number, buf: string }, ... ], where
 * fildes is 0 for stdin (and buf is undefined), 
 * fildes is 1 for stdout, and fildes is 2 for stderr.
 *
 * @param string strace
 *
 * @return object
 */
function parse(strace) {

    // prepare for events
    var syscalls = [];

    // validate strace
    if (!_.isString(strace)) {
        return syscalls;
    }

    // iterate over lines
    var lines = strace.split(/\n/);
    for (var i = 0, matches = []; i < lines.length; i++) {

        // check whether child is waiting for stdin
        if (/^(\[pid\s+\d+\] )?read\(0,/.test(lines[i])) {
            syscalls.push({ fildes: 0 });
        }
        else {

            // check whether child has generated stdout or stdout
            var matches = /^(?:\[pid\s+\d+\] )?write\((1|2), "(|[\s\S]*?[^\\"])"/.exec(lines[i]);
            if (_.isArray(matches)) {
                
                // unescape string
                var s = matches[2]
                 .replace(/(\\|\\\\)([abfnrtv'"])/g, function(match, p1, p2) { // strace escapes '\n' as \n and \n as \\n

                     // replace, e.g., 3-char \\n with 2-char \n
                     if (p1 === '\\\\') {
                          return '\\' + p2;
                     }

                     // replace, e.g., 2-char \n with 1-char '\n'
                     else {
                         switch (p2) {
                             case 'a': return '\a';
                             case 'b': return '\b';
                             case 'f': return '\f';
                             case 'n': return '\n';
                             case 'r': return '\r';
                             case 't': return '\t';
                             case 'v': return '\v';
                             case "'": return "\'";
                             case '"': return '\"';
                             case '?': return '\?';
                         }
                     }
                 })
                 .replace(/\\(\d{1,3})/g, function(match, p1) { // octal, though strace writes ESC as \33 instead of \033
                     return String.fromCharCode(parseInt(p1, 8));
                 })
                 .replace(/\\\\/g, '\\'); // strace escapes \ as \\

                // stdout
                if (matches[1] === '1') {
                    syscalls.push({ fildes: 1, buf: s });
                }

                // stderr
                else if (matches[1] === '2') {
                    syscalls.push({ fildes: 2, buf: s });
                }
            }
        }
    }

    // return strace's syscalls
    return syscalls;
}

/**
 * Runs cmd in sandbox.
 *
 * @param string cmd
 * @param string sandbox
 * @param function callback
 *
 * @return ChildProcess
 */
function run(cmd, sandbox, callback) {

    async.series([

        // validate cmd
        function(callback) {
            if (_.isUndefined(cmd)) {
                callback(error('E_USAGE', 'missing cmd'));
            }
            else if (!_.isString(cmd)) {
                callback(error('E_USAGE', 'invalid cmd'));
            }
            else {
                callback(null);
            }
        },

        // trim cmd to avoid
        // /usr/bin/sandbox: [Errno 17] File exists: '/tmp/.sandbox_tmp_.../sandboxes/xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx
        function(callback) {
            cmd = $.trim(cmd);
            if (cmd.length === 0) {
                callback(error('E_USAGE', 'missing cmd'));
            }
            else {
                callback(null);
            };
        },

        // validate sandbox
        function(callback) {
            if (_.isUndefined(sandbox)) {
                callback(error('E_USAGE', 'missing sandbox'));
            }
            else if (!isId(sandbox)) {
                callback(error('E_USAGE', 'invalid sandbox'));
            }
            else {
                callback(null);
            }
        },

        // ensure sandbox exists
        function(callback) {
            path.exists(path.join(config.SANDBOXES, sandbox), function(exists) {
                if (exists === false) { 
                    callback(error('E_USAGE', 'nonexistent sandbox'));
                }
                else {
                    callback(null);
                }
            });
        },

        // prepare command for child
        function(callback) {

            // homedir
            var homedir = path.join(config.SANDBOXES, sandbox, 'homedir');

            var command = [

             // sudo -u sandbox
             'sudo', '-u', sandbox,

             // nice -n 19
             'nice', '-n', '19',

             // sandbox.sh -t sandbox_t -H homedir -T tmpdir
             path.join(__dirname, 'bin', 'sandbox.sh'), '-t', config.TYPE, '-H', homedir, '-T', path.join(config.SANDBOXES, sandbox, 'tmpdir'),

             // trace reads and writes (-e trace=read,write), child processes (-f), and display first 1MB of strings read/written
             'strace', '-e', 'trace=read,write', '-f', '-s', '1048576',

             // run.sh -H homedir cmd
             path.join(__dirname, 'bin', 'run.sh'), '-H', homedir].concat(cmd.split(/\s+/));

             // TODO: handle quoted args correctly, as might be passed to sed -i -e
             // path.join(__dirname, 'bin', 'run.sh'), '-H', homedir].concat(cmd.match(/'[^']*'|"[^"]*"|\S+/g));

            // spawn child
            callback(null, child_process.spawn(command[0], command.splice(1, command.length - 1), {

             // default $PATH
             env: {
              PATH: [
               '/usr/local/bin',
               '/bin',
               '/usr/bin'
              ].join(':')
             }

            }));
        }
    ], function(err, results) {
        if (err !== null) {
            callback(err);
        }
        else {

            // remember child (in case we need to SIGKILL it later)
            sandboxes[sandbox].child = results.pop();

            // return sandbox's ChildProcess
            callback(null, sandbox, sandboxes[sandbox].child);

        }
    });
}

/**
 * Creates a sandbox with a homedir and tmpdir.
 *
 * @param object|string sandbox
 * @param function callback
 */
function setup(sandbox, callback) {

    // if missing, err
    if (_.isUndefined(sandbox)) {
        callback(error('E_USAGE', 'missing sandbox'));
    }

    // if string, sandbox has already been created
    else if (_.isString(sandbox) && _.isObject(sandboxes[sandbox])) {

        // postpone sandbox's teardown
        sandboxes[sandbox].timeout.touch();

        // return sandbox's ID
        callback(null, sandbox);
    }

    // if object, create sandbox
    else if (_.isObject(sandbox)) {

        // identifier for sandbox
        var id = mkId();

        // create sandbox
        async.series([

            // validate homedir
            function(callback) {
                if (_.isUndefined(sandbox.homedir)) {
                    callback(error('E_USAGE', 'missing sandbox.homedir'));
                }
                else if (!isId(sandbox.homedir)) {
                    callback(error('E_USAGE', 'invalid sandbox.homedir'));
                }
                else {
                    callback(null);
                }
            },


            // ensure upload exists
            function(callback) {

                // check local disk
                path.exists(path.join(config.UPLOADS, sandbox.homedir), function(exists) {
                    if (exists === false) {

                        // check S3
                        if (!_.isUndefined(s3cfg)) {
                            child_process.execFile('s3cmd', ['-c', path.join(__dirname, 'etc', 's3cfg'), 'get', '-r', 's3://' + s3cfg.bucket50 + '/uploads/' + sandbox.homedir, config.UPLOADS], function(err) {
                                if (err !== null) {
                                    callback(error('E_USAGE', 'nonexistent homedir'));
                                }
                                else {
                                    callback(null);
                                }
                            });
                        }
                        else {
                            callback(error('E_USAGE', 'nonexistent homedir'));
                        }

                    }
                    else {
                        callback(null);
                    }
                });

            },

            // mkdir /path/to/sandbox
            function(callback) {

                // mode appears to have no effect in node 0.6.18
                fs.mkdir(path.join(config.SANDBOXES, id), undefined, callback);

            },

            // chmod 0711 /path/to/sandbox
            function(callback) {
                fs.chmod(path.join(config.SANDBOXES, id), 0711, callback);
            },

            // cp -r /path/to/uploads/id /path/to/sandbox/homedir
            function(callback) {
                wrench.copyDirRecursive(path.join(config.UPLOADS, sandbox.homedir), path.join(config.SANDBOXES, id, 'homedir'), callback);
            },

            // chmod 0711 /path/to/sandbox/homedir
            function(callback) {
                fs.chmod(path.join(config.SANDBOXES, id, 'homedir'), 0711, callback);
            },

            // mkdir /path/to/sandbox/tmpdir
            function(callback) {

                // mode appears to have no effect in node 0.6.18
                fs.mkdir(path.join(config.SANDBOXES, id, 'tmpdir'), undefined, callback);

            },

            // chmod 0711 /path/to/sandbox/tmpdir
            function(callback) {
                fs.chmod(path.join(config.SANDBOXES, id, 'tmpdir'), 0711, callback);
            },

            // adduser -g sandboxes -M id
            function(callback) {
                child_process.execFile('sudo', ['adduser', '-d', path.join(config.SANDBOXES, id, 'homedir'), '-g', 'sandboxes', '-m', id], function(err, stdout, stderr) {
                    callback(err);
                });
            },

            // chown -R id:sandboxes /path/to/homedir
            function(callback) {
                child_process.execFile('sudo', ['chown', '-R', [id, ':', 'sandboxes'].join(''), path.join(config.SANDBOXES, id)], function(err, stdout, stderr) {
                    callback(err);
                });
            }

        ], function(err, results) {
            if (err !== null) {

                // clean up after error
                teardown(id);

                // report error
                callback(error('E_UNKNOWN', err.message));

            }
            else {

                // remember sandbox
                sandboxes[id] = {
                 homedir: sandbox.homedir,
                 timeout: _setTimeout(function() {
                     teardown(id);
                 }, config.TIMEOUT * 10)
                };

                // return sandbox's ID
                callback(null, id);
            }
        });
    }
    else {
        callback(error('E_USAGE', 'invalid sandbox'));
    }
}

/**
 * Tears down a sandbox, killing processes, deleting user, and deleting files.
 *
 * @param string sandbox
 * @param function callback
 */
function teardown(sandbox, callback) {

    // validate sandbox
    if (!isId(sandbox)) {
        if (_.isFunction(callback)) {
            callback(error('E_USAGE', 'invalid sandbox'));
        }
        return;
    }
    else if (!_.isObject(sandboxes[sandbox])) {
        if (_.isFunction(callback)) {
            callback(null); // don't worry if sandbox was already torn down
        }
        return;
    }

    // tear down sandbox
    async.series([

        // kill sandbox's process(es)
        function(callback) {
            kill(sandbox, function(err) {
                callback(null); // don't worry if no processes were killed
            });
        },

        // userdel -r sandbox
        function(callback) {
            child_process.execFile('sudo', ['userdel', '-f', '-r', sandbox], function(err, stdout, stderr) {
                callback(null); // don't worry if no user was deleted
            });
        },

        // rm -rf /path/to/sandbox
        function(callback) {
            child_process.execFile('sudo', ['rm', '-rf', path.join(config.SANDBOXES, sandbox)], function(err, stdout, stderr) {
                callback(null); // don't worry if no sandbox was removed
            });
        }

    ], function(err) {
        delete sandboxes[sandbox];
        if (_.isFunction(callback)) {
            callback(err);
        }
    });
}
